// Adven.cpp

#include <PalmOS.h>
#include "Adven.h"
#include "Resource.h"
#include "Utility.h"
#include "advent.h"
#include "advdef.h"
#include "advword.h"	/* definition of "word" array	*/
#include "advcave.h"	/* definition of "cave" array	*/

/***********************************************************************
 *
 *   Entry Points
 *
 ***********************************************************************/
UInt32	PilotMain (UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags);

/***********************************************************************
 *
 *   Global variables
 *
 ***********************************************************************/

DmOpenRef AppDB = NULL;

// The following global variable are saved in preferences
static FontID NoteFont = stdFont;
static Boolean bStartNew = true;

/***********************************************************************
 *
 * Internal Functions
 *
 ***********************************************************************/

static UInt32 AddrPilotMain (UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags);

/*=========================================================================*/

static void ShowAboutBox()
{
	FormPtr dlg = FrmInitForm (AboutBoxForm);

	UInt16 but = FrmDoDialog (dlg);
	FrmDeleteForm (dlg);

//	if (but == AboutBoxRegisterButButton)
//		FrmPopupForm (RegisterForm);
}

/*-------------------------------------------------------------------------*/

FieldPtr GetObjectPtr (UInt16 objectID)
{
	FormPtr frm = FrmGetActiveForm ();
	return (FieldPtr)FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, objectID));
}

/*-------------------------------------------------------------------------*/
//	Initialization of adventure play variables

static void scanint (int* pi, char* str)
{

	while (*str) {
		*pi = StrAToI (str); pi++;
		while (*str++ != ',')	/* advance str pointer	*/
			;
	}
	return;
}

static void initplay()
{
	turns = 0;

	/* initialize location status array */
	MemSet(cond, (sizeof(int))*MAXLOC, 0);
	scanint(&cond[1], "5,1,5,5,1,1,5,17,1,1,");
	scanint(&cond[13], "32,0,0,2,0,0,64,2,");
	scanint(&cond[21], "2,2,0,6,0,2,");
	scanint(&cond[31], "2,2,0,0,0,0,0,4,0,2,");
	scanint(&cond[42], "128,128,128,128,136,136,136,128,128,");
	scanint(&cond[51], "128,128,136,128,136,0,8,0,2,");
	scanint(&cond[79], "2,128,128,136,0,0,8,136,128,0,2,2,");
	scanint(&cond[95], "4,0,0,0,0,1,");
	scanint(&cond[113], "4,0,1,1,");
	scanint(&cond[122], "8,8,8,8,8,8,8,8,8,");

	/* initialize object locations */
	MemSet(place, (sizeof(int))*MAXOBJ, 0);
	scanint(&place[1], "3,3,8,10,11,0,14,13,94,96,");
	scanint(&place[11], "19,17,101,103,0,106,0,0,3,3,");
	scanint(&place[23], "109,25,23,111,35,0,97,");
	scanint(&place[31], "119,117,117,0,130,0,126,140,0,96,");
	scanint(&place[50], "18,27,28,29,30,");
	scanint(&place[56], "92,95,97,100,101,0,119,127,130,");

	/* initialize second (fixed) locations */
	MemSet(fixed, (sizeof(int))*MAXOBJ, 0);
	scanint(&fixed[3], "9,0,0,0,15,0,-1,");
	scanint(&fixed[11], "-1,27,-1,0,0,0,-1,");
	scanint(&fixed[23], "-1,-1,67,-1,110,0,-1,-1,");
	scanint(&fixed[31], "121,122,122,0,-1,-1,-1,-1,0,-1,");
	scanint(&fixed[62], "121,-1,");

	/* initialize default verb messages */
	scanint(actmsg, "0,24,29,0,33,0,33,38,38,42,14,");
	scanint(&actmsg[11], "43,110,29,110,73,75,29,13,59,59,");
	scanint(&actmsg[21], "174,109,67,13,147,155,195,146,110,13,13,");

	/* initialize various flags and other variables */
	MemSet(visited, (sizeof(int))*MAXLOC, 0);
	MemSet(prop, (sizeof(int))*MAXOBJ, 0);
	MemSet(&prop[50], (sizeof(int))*(MAXOBJ-50), 0xff);
	wzdark = closed = closing = holding = detail = 0;
	limit = 1000; // was 330 if no rules requested
	tally = 15;
	tally2 = 0;
	newloc = 3;
	loc = oldloc = oldloc2 = 1;
	knfloc = 0;
	chloc = 114;
	chloc2 = 140;
/*	dloc[DWARFMAX-1] = chloc				*/
	scanint(dloc, "0,19,27,33,44,64,114,");
	scanint(odloc, "0,0,0,0,0,0,0,");
	dkill = 0;
	scanint(dseen, "0,0,0,0,0,0,0,");
	clock1 = 30;
	clock2 = 50;
	panic = 0;
	bonus = 0;
	numdie = 0;
	daltloc = 18;
	lmwarn = 0;
	foobar = 0;
	dflag = 0;
	gaveup = 0;
	saveflg = 0;
	dbugflg = 0;
	return;
}

/*-------------------------------------------------------------------------*/

static void AppLoadPrefs (MyInfoPtr appInfoPtr)
{
	// Read the preferences / saved-state information.
	MyPreferenceType prefs;
	UInt16 prefsSize = sizeof (MyPreferenceType);
	Int16 prefsVersion = PrefGetAppPreferences (CreatorID, MyPrefID, &prefs, &prefsSize, true);
	if (prefsVersion > MyPrefVersionNum) {
		prefsVersion = noPreferenceFound;
	}
	if (prefsVersion > noPreferenceFound) {
		NoteFont = prefs.noteFont;
		bStartNew = prefs.bStartNew;
		// add more prefs here
	}
	
	MemPtrUnlock(appInfoPtr);
}

/*-------------------------------------------------------------------------*/

static void AppSavePrefs(void)
{
	MyPreferenceType prefs;

	// Write the preferences / saved-state information.
	prefs.noteFont = NoteFont;

	if (terminated) {
		prefs.bStartNew = bStartNew;
	} else {
		prefs.bStartNew = false;
		DBSaveGame ("Default");
	}

	// Write the state information.
	PrefSetAppPreferences (CreatorID, MyPrefID, MyPrefVersionNum, &prefs, 
		sizeof (MyPreferenceType), true);
}

/*-------------------------------------------------------------------------*/

static Err StartApplication(void)
{
	// Find the application's data file.  If it doesn't exist create it.
	Err err = GetAppDatabase (&AppDB, dmModeReadWrite);
	if (err) return err;

	MyInfoPtr appInfoPtr = MyInfoGetPtr(AppDB);
	ErrFatalDisplayIf(appInfoPtr == NULL, "Missing app info block");
	
	// Load the application preferences
	AppLoadPrefs(appInfoPtr);

	// Mask off the key to avoid repeat keys causing clicking sounds
	KeySetMask(~KeyCurrentState());

	return 0;
}

/*-------------------------------------------------------------------------*/
// This routine close the application's database
// and save the current state of the application.

static void StopApplication(void)
{
	// Write the preferences / saved-state information.
	AppSavePrefs();

	// Send a frmSave event to all the open forms.
	FrmSaveAllForms ();

	// Close all the open forms.
	FrmCloseAllForms ();

	// Close the application's data file.
	DmCloseDatabase (AppDB);
}

/*-------------------------------------------------------------------------*/
// Leaves the current popup form and returns to the prior one.

static void LeaveForm()
{
	FormPtr frm = FrmGetActiveForm();
	FrmEraseForm (frm);
	FrmDeleteForm (frm);
	FrmSetActiveForm (FrmGetFirstForm());
}

/*-------------------------------------------------------------------------*/

static void MainViewUpdateScrollBar()
{
	UInt16 scrollPos, textHeight, fieldHeight;
	FieldPtr fld = (FieldPtr)GetObjectPtr (MainScreenField);
	FldGetScrollValues (fld, &scrollPos, &textHeight, &fieldHeight);

	Int16 maxValue = 0;
	if (textHeight > fieldHeight) {
		maxValue = (textHeight - fieldHeight) + FldGetNumberOfBlankLines (fld);
	} else if (scrollPos)
		maxValue = scrollPos;

	ScrollBarPtr bar = (ScrollBarPtr)GetObjectPtr (MainScrollbarScrollBar);
	SclSetScrollBar (bar, scrollPos, 0, maxValue, fieldHeight-1);
}

/*-------------------------------------------------------------------------*/

static void MainViewScroll (Int16 linesToScroll, Boolean updateScrollbar)
{
	FieldPtr fld = GetObjectPtr (MainScreenField);
	UInt16 blankLines = FldGetNumberOfBlankLines (fld);

	if (linesToScroll < 0)
		FldScrollField (fld, -linesToScroll, winUp);
	else if (linesToScroll > 0)
		FldScrollField (fld, linesToScroll, winDown);
		
	// If there were blank lines visible at the end of the field
	// then we need to update the scroll bar.
	if (blankLines && linesToScroll < 0 || updateScrollbar)
		MainViewUpdateScrollBar();
}

/*-------------------------------------------------------------------------*/

static void MainViewPageScroll (WinDirectionType direction)
{
	FieldPtr fld = (FieldPtr)GetObjectPtr (MainScreenField);
	if (FldScrollable (fld, direction)) {
		Int16 linesToScroll = FldGetVisibleLines (fld) - 1;
		if (direction == winUp)
			linesToScroll = -linesToScroll;
		
		MainViewScroll (linesToScroll, true);
	}
}

/*-------------------------------------------------------------------------*/

void FldReplaceText (FieldPtr fld, LPCSTR pText)
{
	// first create a new handle
	MemHandle hNew = MemHandleNew (StrLen (pText) + 1);
	if (hNew == NULL) return; // out of mem
	Char* ptr = (Char*)MemHandleLock (hNew);
	StrCopy (ptr, pText);
	MemHandleUnlock (hNew);

	// give hnd to the field
	MemHandle hOldTxt = FldGetTextHandle (fld);
	if (fld->attr.editable)
		FldSetSelection (fld, 0, 0);
	FldSetTextHandle (fld, hNew);

	// now delete old handle
	if (hOldTxt)
		MemHandleFree (hOldTxt);
		
	// last but not least: draw!
	FldDrawField (fld);
}

/*-------------------------------------------------------------------------*/

static void OutputText()
{
	FieldPtr fld = GetObjectPtr (MainScreenField);
	MemHandle hOldTxt = FldGetTextHandle (fld);

	// give ScreenData to the field
	FldSetTextHandle (fld, ScreenData);
	ScreenData = NULL; // it's not mine any more
	FldDrawField (fld);
	MainViewUpdateScrollBar();

	// now delete old handle
	if (hOldTxt)
		MemHandleFree (hOldTxt);

	// pop the score onto the screen
	char AsciiScore[32];
	StrIToA (AsciiScore, score(false));
	FldReplaceText (GetObjectPtr (MainScoreField), AsciiScore);
}

/*-------------------------------------------------------------------------*/

void StartNewGame (Int16 index)
{
	// prepare to play
	initplay();
	if (index >= 0) {
		if (DBRestoreGame (index) != 0) {
			initplay(); // reset on error
			char buff[31]; StrIToA (buff, index);
			FrmCustomAlert (DBErrorAlert, buff, NULL, NULL);
		}
	}
	saveflg = 0;
	terminated = 0;
	srandom(511);				/* seed random	*/
	
	ProcessCommand (NULL);
	if (ScreenData == NULL) { // usually after restore
		describe();
		if (!dark())
			descitem();
	}

	OutputText();
}

/*-------------------------------------------------------------------------*/

static void MainHandleOK()
{
	FieldPtr pFld = GetObjectPtr (MainCommandField);
	char* pCmd = FldGetTextPtr (pFld);
	if (pCmd == NULL) return;
	ProcessCommand (pCmd);
	OutputText();
	FldSetSelection (pFld, 0, FldGetTextLength (pFld));
}

/*-------------------------------------------------------------------------*/

static void MainHandlePopSelect (EventPtr event)
{
	ListPtr pList = event->data.popSelect.listP;
	Char* pText = LstGetSelectionText (pList, event->data.popSelect.selection);

	// pop the text into the command field 
	FldReplaceText (GetObjectPtr (MainCommandField), pText);

	// if the last character is not a space hit OK
	if (*pText && pText[StrLen(pText)-1] != ' ')
		MainHandleOK();
}

/*-------------------------------------------------------------------------*/

static Boolean MainHandlePen (EventPtr event)
{
	struct MapObj {
		RectangleType rect;
		LPCSTR cmd;
	} map[] = {
		{{6,0,5,6}, "north"},
		{{11,5,6,5}, "east"},
		{{6,9,5,6}, "south"},
		{{0,5,6,5}, "west"},
		{{11,0,6,5}, "ne"},
		{{0,0,6,5}, "nw"},
		{{11,10,6,5}, "se"},
		{{0,10,6,5}, "sw"},
		{{18,0,8,7}, "up"},
		{{18,8,8,7}, "down"},
	};

	FormPtr frm = FrmGetActiveForm();
	Coord x = event->screenX;
	Coord y = event->screenY;
	RectangleType rc;
	FrmGetObjectBounds (frm, FrmGetObjectIndex (frm, MainUpBitMap), &rc);

	// clear last invert if there was one
	static RectangleType rcLastBut;
	if (rcLastBut.extent.x != 0) {
		WinInvertRectangle (&rcLastBut, 0);
		rcLastBut.extent.x = 0; // clear
	}

	if (RctPtInRectangle (x, y, &rc)) {
		// user clicked the direction map
		x -= rc.topLeft.x; y -= rc.topLeft.y; // make relative to map
		for (int i = 0; i < sizeof(map)/sizeof(MapObj); i++) {
			if (RctPtInRectangle (x, y, &map[i].rect)) {
				// a hot spot was hit
				if (event->penDown) {
					rcLastBut = map[i].rect;
					RctOffsetRectangle (&rcLastBut, rc.topLeft.x, rc.topLeft.y);
					WinInvertRectangle (&rcLastBut, 0);
				} else {
					FldReplaceText (GetObjectPtr (MainCommandField), map[i].cmd);
					MainHandleOK();
				}
				break;
			}
		}
		return true;
	}
	
	return false;
}

/*-------------------------------------------------------------------------*/

static FontID SelectFont (FontID currFontID)
{
	UInt16 formID = FrmGetFormId (FrmGetActiveForm());

	// Call the OS font selector to get the id of a font.
	FontID fontID = FontSelect (currFontID);

	if (fontID != currFontID)
		FrmUpdateForm (formID, updateFontChanged);

	return fontID;
}

/*-------------------------------------------------------------------------*/

static Boolean MainDoCommand (UInt16 command)
{
	switch (command) {
	case noteFontCmd: // for Palm 3.0
	case newNoteFontCmd:
		NoteFont = SelectFont (NoteFont);
		break; 

	case MainViewGameRules:
		rspeak(1);
		OutputText();
		return true;

	case MainViewGameRestart:
		StartNewGame (-1);
		return true;

	case MainViewGameSave:
		FrmPopupForm (SaveGameForm);
		return true;

	case MainViewGameRestore:
		FrmPopupForm (GameListForm);
		return true;

	case MainViewOptionsFont:
		MenuEraseStatus (0);
		NoteFont = SelectFont (NoteFont);
		return true;

	case MainViewOptionsAboutAdventure:
		MenuEraseStatus (0);
		ShowAboutBox();
		return true;
	}

	return false;
}

/*-------------------------------------------------------------------------*/
// This is the main even handler for Main Form

static Boolean MainHandleEvent (EventPtr event)
{
	Boolean handled = false;

	switch (event->eType) {
	case keyDownEvent:
		if (EvtKeydownIsVirtual(event)) {
			if (event->data.keyDown.chr == vchrPageUp) {
				MainViewPageScroll (winUp);
				handled = true;
			} else if (event->data.keyDown.chr == vchrPageDown) {
				MainViewPageScroll (winDown);
				handled = true;
			}
		}
		break;

	case ctlSelectEvent:
		switch (event->data.ctlSelect.controlID) {
		case MainOKButton:
			MainHandleOK();
			handled = true;
			break;
		
		default:
			break;
		}
		break;

	case popSelectEvent:
		MainHandlePopSelect (event);
		handled = true;
		break;

	case menuEvent:
		return MainDoCommand (event->data.menu.itemID);

	case penDownEvent:
	case penUpEvent:
	case penMoveEvent:
		handled = MainHandlePen (event);
		break;

	case frmOpenEvent:
	{
		// Set the font used in the memo field.
		FldSetFont (GetObjectPtr (MainScreenField), NoteFont);
		FormPtr frm = FrmGetActiveForm();
		FrmDrawForm (frm);
		FrmSetFocus (frm, FrmGetObjectIndex (frm, MainCommandField));

		// remove the autoshift
		FieldAttrType attr;
		FieldPtr fld = GetObjectPtr (MainCommandField);
		FldGetAttributes (fld, &attr);
		attr.autoShift = false;
		FldSetAttributes (fld, &attr);

		// let's go
		StartNewGame (bStartNew ? -1 : DBFindDefaultGame());
		handled = true;
		break;
	}

	case frmUpdateEvent:
		if (event->data.frmUpdate.updateCode & updateFontChanged) {
			FldSetFont (GetObjectPtr (MainScreenField), NoteFont);
			MainViewUpdateScrollBar();
			handled = true;
		}
		break;

	case sclRepeatEvent:
		MainViewScroll (event->data.sclRepeat.newValue - 
			event->data.sclRepeat.value, false);
		break;
	}

	return handled;
}

/*-------------------------------------------------------------------------*/

static Boolean ApplicationHandleEvent (EventPtr event)
{
	if (event->eType == frmLoadEvent) {
		
		// Load the form resource.
		UInt16 formId = event->data.frmLoad.formID;
		FormPtr frm = FrmInitForm (formId);
		FrmSetActiveForm (frm); 	  
		
		// Set the event handler for the form.  The handler of the currently
		// active form is called by FrmHandleEvent each time it receives an
		// event.
		switch (formId) {
		case MainForm:
			FrmSetEventHandler(frm, MainHandleEvent);
			break;
	
		case SaveGameForm:
		case GameListForm:
			FrmSetEventHandler(frm, GameListHandleEvent);
			break;
	
		default:
			ErrNonFatalDisplay("Invalid Form Load Event");
			break;
		}
	
		return true;
	}

	return false;
}

/*-------------------------------------------------------------------------*/
// Main Event Loop

static void EventLoop (void)
{
	UInt16 error;
	EventType event;

	do {
		EvtGetEvent (&event, evtWaitForever);
		if (!SysHandleEvent (&event) &&
			!MenuHandleEvent (0, &event, &error) &&
			!ApplicationHandleEvent (&event))
			FrmDispatchEvent (&event); 
	} while (event.eType != appStopEvent);
}

/*-------------------------------------------------------------------------*/
// Here's where it all begins...

UInt32 PilotMain (UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	return AddrPilotMain(cmd, cmdPBP, launchFlags);
}

/*=========================================================================*/

static UInt32 AddrPilotMain (UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err error = errNone;
	DmOpenRef dbP;
			 
	switch (cmd) {
	case sysAppLaunchCmdNormalLaunch:
		error = StartApplication();
		if (error) return (error);
		
		FrmGotoForm (MainForm);
		EventLoop();

		StopApplication();
		break;


	// Launch code sent to running app before sysAppLaunchCmdFind
	// or other action codes that will cause data searches or manipulation.
	case sysAppLaunchCmdSaveData:
		FrmSaveAllForms();
		break;

		
	// We are requested to initialize an empty database (by sync app).
	case sysAppLaunchCmdInitDatabase:
		AppLaunchCmdDatabaseInit (((SysAppLaunchCmdInitDatabaseType*)cmdPBP)->dbP);
		break;
	
	
	// This launch code is sent after the system is reset.  We use this time
	// to create our default database.	If there is no default database image,
	// then we create an empty database.
	case sysAppLaunchCmdSystemReset:
		if (((SysAppLaunchCmdSystemResetType*)cmdPBP)->createDefaultDB) {
			
			// Attempt to get our default data image and create our database.
			MemHandle resH = DmGet1Resource(sysResTDefaultDB, sysResIDDefaultDB);
			if (resH) {
				error = DmCreateDatabaseFromImage(MemHandleLock(resH));
				
				if (!error) {
					MemHandleUnlock(resH);
					DmReleaseResource(resH);
				
					// Set the backup bit on the new database.
					SetDBAttrBits (NULL, dmHdrAttrBackup);
				}
			}

			// If there is no default data, or we had a problem creating it,
			// then attempt to create an empty database.
			if (!resH || error) {
				error = GetAppDatabase (&dbP, dmModeReadWrite);

				if (!error)
					DmCloseDatabase(dbP);
			}
		}
		break;
	}
	
	return error;
}

