// AppData.cpp

#include <PalmOS.h>
#include "Adven.h"
#include "Advent.h"
#include "advdec.h"

/*=========================================================================*/

MRecord::MRecord (UInt16 index)
{
	MemHandle hRecord = DmGetRecord (AppDB, index);
	if (hRecord) {
		Char* pData = (Char*)MemHandleLock (hRecord);
		m_pData = (Char*)MemPtrNew (StrLen (pData) + 1);
		StrCopy (m_pData, pData);
		MemHandleUnlock (hRecord);
		DmReleaseRecord (AppDB, index, false);
	} else
		m_pData = NULL;
}

/*-------------------------------------------------------------------------*/

void MRecord::Empty()
{
	if (m_pData)
		MemPtrFree (m_pData);
	m_pData = NULL;
}

/*=========================================================================*/

void SetDBAttrBits (DmOpenRef dbP, UInt16 attrBits)
{
	DmOpenRef localDBP;
	LocalID dbID;
	UInt16 cardNo;
	UInt16 attributes;

	// Open database if necessary. If it doesn't exist, simply exit (don't create it).
	if (dbP == NULL) {
		localDBP = DmOpenDatabaseByTypeCreator (MyDBType, CreatorID, dmModeReadWrite);
		if (localDBP == NULL)  return;
	} else
		localDBP = dbP;
	
	// now set the backup bit on localDBP
	DmOpenDatabaseInfo(localDBP, &dbID, NULL, NULL, &cardNo, NULL);
	DmDatabaseInfo(cardNo, dbID, NULL, &attributes, NULL, NULL,
				NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	attributes |= attrBits;
	DmSetDatabaseInfo(cardNo, dbID, NULL, &attributes, NULL, NULL,
				NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	
	// close database if necessary
	if (dbP == NULL)
		DmCloseDatabase (localDBP);
}

/*-------------------------------------------------------------------------*/

MyInfoPtr MyInfoGetPtr (DmOpenRef dbP)
{
	UInt16  cardNo;
	LocalID dbID;
	LocalID appInfoID;
	
	if (DmOpenDatabaseInfo (dbP, &dbID, NULL, NULL, &cardNo, NULL))
		return NULL;
	if (DmDatabaseInfo (cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
		return NULL;

	if (appInfoID == NULL)
		return NULL;

	return (MyInfoPtr)MemLocalIDToLockedPtr (appInfoID, cardNo);
}	 

/*-------------------------------------------------------------------------*/

UInt16 DBFindDefaultGame()
{
	UInt16 nRec = DmNumRecords (AppDB);
	for (UInt16 i = 0; i < nRec; i++) {
		MRecord rec (i);
		if (StrCompare (rec.Name(), "Default") == 0)
			return i;
	}
	return -1;
}

/*-------------------------------------------------------------------------*/

static UInt16 FindGame (Char* pName)
{
	UInt16 index = 0;
	for (;;) { // for ever
		MemHandle hRec = DmQueryNextInCategory (AppDB, &index, dmAllCategories);
		if (!hRec) return -1; // not found
		
		Char* pData = (Char*)MemHandleLock (hRec);
		Boolean bFound = StrCompare (pData, pName) == 0;
		MemHandleUnlock (hRec);
		if (bFound)
			return index;  // found it

		index++;
	}

	return -1;
}

/*-------------------------------------------------------------------------*/

Err DBSaveGame (Char* pName)
{
	if (pName == NULL)
		pName = "Default";

	// copy the game to a new chunk of dynamic memory
	MStreamOut stm (1270);
	stm.Put (pName);
	stm.Put (turns);
	stm.Put (loc);
	stm.Put (oldloc);
	stm.Put (oldloc2);
	stm.Put (newloc);
	int i;
	for(i=0;i<MAXLOC;i++) stm.Put (cond[i]);
	for(i=0;i<MAXOBJ;i++) stm.Put (place[i]);
	for(i=0;i<MAXOBJ;i++) stm.Put (fixed[i]);
	for(i=0;i<MAXLOC;i++) stm.Put (visited[i]);
	for(i=0;i<MAXOBJ;i++) stm.Put (prop[i]);
	stm.Put (tally);
	stm.Put (tally2);
	stm.Put (limit);
	stm.Put (lmwarn);
	stm.Put (wzdark);
	stm.Put (closing);
	stm.Put (closed);
	stm.Put (holding);
	stm.Put (detail);
	stm.Put (knfloc);
	stm.Put (clock1);
	stm.Put (clock2);
	stm.Put (panic);
	for(i=0;i<DWARFMAX;i++) stm.Put (dloc[i]);
	stm.Put (dflag);
	for(i=0;i<DWARFMAX;i++) stm.Put (dseen[i]);
	for(i=0;i<DWARFMAX;i++) stm.Put (odloc[i]);
	stm.Put (daltloc);
	stm.Put (dkill);
	stm.Put (chloc);
	stm.Put (chloc2);
	stm.Put (bonus);
	stm.Put (numdie);
	stm.Put (object1);
	stm.Put (gaveup);
	stm.Put (foobar);
	stm.Put (saveflg);
	stm.Put (dbugflg);
	if (stm.Error()) 
		return dmErrMemError;
		
	// make a new chunk with the correct size
	MemHandle hRec = DmNewHandle (AppDB, stm.GetLength());
	if (hRec == NULL) return dmErrMemError;
		
	// Copy the data 
	MemPtr pData = MemHandleLock (stm.GetHandle()); 
	MemPtr pRec = MemHandleLock (hRec);
	Err err = DmWrite (pRec, 0, pData, stm.GetLength());
	MemPtrUnlock (pRec);
	MemPtrUnlock (pData);

	UInt16 index = FindGame (pName);
	if (index != (UInt16)-1) {
		MemHandle hOld;
		err = DmAttachRecord (AppDB, &index, hRec, &hOld);
		MemHandleFree (hOld);
	} else {
		index = 0; // insert on the top
		err = DmAttachRecord (AppDB, &index, hRec, NULL);
	}
	if (err) 
		MemHandleFree (hRec);

	return err;
}

/*-------------------------------------------------------------------------*/

Err DBRestoreGame (UInt16 index)
{
	Err err = 0;
	MemHandle hRec = DmGetRecord (AppDB, index);
	MemPtr pData = MemHandleLock (hRec);
	if (pData == NULL) return dmErrNotValidRecord;

	// suck it all back in
	MStreamIn stm ((const char*)pData, MemHandleSize(hRec));
	stm.EatData (StrLen (stm.GetRaw()) + 1); // eat the name
	turns = stm.GetW();
	loc = stm.GetW();
	oldloc = stm.GetW();
	oldloc2 = stm.GetW();
	newloc = stm.GetW();
	int i;
	for(i=0;i<MAXLOC;i++) cond[i] = stm.GetW();
	for(i=0;i<MAXOBJ;i++) place[i] = stm.GetW();
	for(i=0;i<MAXOBJ;i++) fixed[i] = stm.GetW();
	for(i=0;i<MAXLOC;i++) visited[i] = stm.GetW();
	for(i=0;i<MAXOBJ;i++) prop[i] = stm.GetW();
	tally = stm.GetW();
	tally2 = stm.GetW();
	limit = stm.GetW();
	lmwarn = stm.GetW();
	wzdark = stm.GetW();
	closing = stm.GetW();
	closed = stm.GetW();
	holding = stm.GetW();
	detail = stm.GetW();
	knfloc = stm.GetW();
	clock1 = stm.GetW();
	clock2 = stm.GetW();
	panic = stm.GetW();
	for(i=0;i<DWARFMAX;i++) dloc[i] = stm.GetW();
	dflag = stm.GetW();
	for(i=0;i<DWARFMAX;i++) dseen[i] = stm.GetW();
	for(i=0;i<DWARFMAX;i++) odloc[i] = stm.GetW();
	daltloc = stm.GetW();
	dkill = stm.GetW();
	chloc = stm.GetW();
	chloc2 = stm.GetW();
	bonus = stm.GetW();
	numdie = stm.GetW();
	object1 = stm.GetW();
	gaveup = stm.GetW();
	foobar = stm.GetW();
	saveflg = stm.GetW();
	dbugflg = stm.GetW();
	if (stm.Error())
		err = dmErrNotValidRecord;

	MemHandleUnlock (hRec);
	DmReleaseRecord (AppDB, index, false);
	return err;
}

/*-------------------------------------------------------------------------*/

static Err MyInfoInit(DmOpenRef dbP)
{
	UInt16			cardNo;
	LocalID			dbID;
	LocalID			appInfoID;
	MemHandle			h;

	MyInfoPtr appInfoP = MyInfoGetPtr (dbP);

	// If there isn't an AddrApplicationInfo make space for one
	if (appInfoP == NULL) {
		if (DmOpenDatabaseInfo(dbP, &dbID, NULL, NULL, &cardNo, NULL))
			return dmErrInvalidParam;
		if (DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
			return dmErrInvalidParam;

		h = DmNewHandle (dbP, sizeof(MyInfoType));
		if (!h) return dmErrMemError;

		appInfoID = (LocalID)MemHandleToLocalID( h);
		DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL);

		appInfoP = (MyInfoPtr)MemHandleLock(h);
	}

	// Allocate & Clear the app info
	MyInfoPtr pDefAppInfo = (MyInfoPtr)MemPtrNew(sizeof(MyInfoType));
	if (pDefAppInfo == NULL)
		return dmErrMemError;

	MemSet(pDefAppInfo, sizeof(MyInfoType), 0);

	// copy in the defaults and free the default app info
	DmWrite (appInfoP, 0, pDefAppInfo,	sizeof(MyInfoType));
	MemPtrFree (pDefAppInfo);

	// Unlock
	MemPtrUnlock(appInfoP);

	return 0;
}

/*-------------------------------------------------------------------------*/

Err GetAppDatabase (DmOpenRef *dbPP, UInt16 mode)
{
	*dbPP = NULL;

	// Find the application's data file.  If it doesn't exist create it.
	DmOpenRef dbP = DmOpenDatabaseByTypeCreator (MyDBType, CreatorID, mode);
	if (!dbP) {
		Err error = DmCreateDatabase (0, MyDBName, CreatorID, MyDBType, false);
		if (error) return error;
		
		dbP = DmOpenDatabaseByTypeCreator (MyDBType, CreatorID, mode);
		if (!dbP) return (1);

		// Set the backup bit.	This is to aid syncs with non Palm software.
		SetDBAttrBits(dbP, dmHdrAttrBackup);
		
		error = MyInfoInit (dbP);
		if (error) {
			UInt16 cardNo;
			LocalID dbID;
			DmOpenDatabaseInfo (dbP, &dbID, NULL, NULL, &cardNo, NULL);
			DmCloseDatabase (dbP);
			DmDeleteDatabase (cardNo, dbID);
			return error;
		}
	}
	
	*dbPP = dbP;
	return 0;
}

/*-------------------------------------------------------------------------*/

Boolean AppLaunchCmdDatabaseInit (DmOpenRef dbP)
{
	if (!dbP) return false;

	// Set the backup bit.	This is to aid syncs with non Palm software.
	SetDBAttrBits (dbP, dmHdrAttrBackup);

	// Initialize the database's app info block
	Err err = MyInfoInit (dbP);
	if (err) return false;

	return true;
}

