
/*	program DATABASE.C					*\
\*	WARNING: "advent.c" allocates GLOBAL storage space by	*\
\*		including "advdef.h".				*\
\*		All other modules use "advdec.h".		*/


#include <PalmOS.h>
#include <stdarg.h>
#include "Resource.h"
#include "advent.h"
#include "advdec.h"

#define ID_TABLE1	3000
#define ID_TABLE2	3500
#define ID_TABLE3	4000
#define ID_TABLE4	4500

/*-------------------------------------------------------------------------*/

int printf (const char *fmtStr, ...)
{
	va_list args;
	va_start(args, fmtStr);
	char buff[160];
	StrVPrintF (buff, fmtStr, args);
	va_end(args);
	ScreenAdd (buff);
	return 0;
}

/*-------------------------------------------------------------------------*/
// same as strrchr
static char* rindex (char* str, int ch)
{
	char* ptr = str + StrLen (str) - 1;
	while (ptr >= str) {
		if (*ptr == ch) return ptr; // found
		ptr--;
	}
	return NULL;
}

/*
	Routine to fill travel array for a given location
*/
void gettrav (int loc)
{
	int	i;
	long	t;
	char	atrav[256], *aptr;

	StrCopy (atrav, cave[loc - 1]);
	while ((aptr = rindex(atrav, ',')) != 0)
		*aptr = '\0';		/* terminate substring	*/
	aptr = &atrav[0];
	for (i = 0; i < MAXTRAV; ++i) {
		t = StrAToI (aptr); 	/* convert to long int	*/
		travel[i].tcond = (t % 1000);
		t /= 1000;
		travel[i].tverb = (t % 1000);
		t /= 1000;
		travel[i].tdest = (t % 1000);
		while (*aptr++)
			;		/* to next substring	*/
		if (!(*aptr)) {
		    travel[++i].tdest = -1;/* end of array	*/
		    if (dbugflg)
			for (i = 0; i < MAXTRAV; ++i)
				printf("cave[%d] = %d %d %d\n", \
				loc, travel[i].tdest, \
				travel[i].tverb, travel[i].tcond);
		    return;		/* terminate for loop	*/
		}
	}
	bug(33);
}

/*
	Function to scan a file up to a specified
	point and either print or return a string.
*/

static void PrintTo (LPCSTR pText, char uptoc)
{
	LPCSTR ptr;
	if (uptoc == 0 || !(ptr = StrChr (pText, uptoc)))
		ScreenAdd (pText);
	else
		ScreenAdd (pText, ptr-pText);
}

/*
	Function to read a file skipping
	a given character a specified number
	of times, with or without repositioning
	the file.
*/
static LPCSTR rdskip (LPCSTR pText, char skipc, int n)
{
	while (n--) {
		LPCSTR ptr = StrChr (pText, skipc);
		if (ptr == NULL)
			bug(32);
		pText = ptr + 1;
	}
	return pText;
}

/*
	Routine to request a yes or no answer to a question.
*/
int yes (int msg1, int msg2, int msg3)
{
	if (msg1 == 0) return 0;
	MResStr pStr (ID_TABLE4+msg1);

	if (FrmCustomAlert(ConfirmAlert, pStr, NULL, NULL) != ConfirmYes) {
		if (msg3)
			rspeak(msg3);
		return(0);
	}
	if (msg2)
		rspeak(msg2);
	return(1);
}

/*
	Print a location description from "advent4.txt"
*/
void rspeak (int msg)
{
	if (msg == 54)
		printf("ok.\n");
	else {
		PrintTo (MResStr (ID_TABLE4+msg), 0);
	}
	return;
}

/*
	Print an item message for a given state from "advent3.txt"
*/
void pspeak (int item, int state)
{
	MResStr pStr (ID_TABLE3+item);
	LPCSTR ptr = rdskip (pStr, '/', state+2);
	PrintTo (ptr, '/');
}

/*
	Print a long location description from "advent1.txt"
*/
void desclg (int loc)
{
	PrintTo (MResStr (ID_TABLE1+loc), 0);
}

/*
	Print a short location description from "advent2.txt"
*/
void descsh (int loc)
{
	PrintTo (MResStr (ID_TABLE2+loc), 0);
}

static int binary (char* w, struct wac wctable[], int maxwc)
{
	int	lo, mid, hi, check;

	lo = 0;
	hi = maxwc - 1;
	while (lo <= hi) {
		mid = (lo + hi) / 2;
		if ((check = StrCompare (w, wctable[mid].aword)) < 0)
			hi = mid - 1;
		else if (check > 0)
			lo = mid + 1;
		else
			return(mid);
	}
	return(-1);
}

/*-------------------------------------------------------------------------*/

/*
	look-up vocabulary word in lex-ordered table.  words may have
	two entries with different codes. if minimum acceptable value
	= 0, then return minimum of different codes.  last word CANNOT
	have two entries(due to binary sort).
	word is the word to look up.
	val  is the minimum acceptable value,
		if != 0 return %1000
*/
short vocab (char* word, int val)
{
	int	v1, v2;

	if ((v1 = binary(word, wc, MAXWC)) >= 0) {
		v2 = binary(word, wc, MAXWC-1);
		if (v2 < 0)
			v2 = v1;
		if (!val)
			return(wc[v1].acode < wc[v2].acode\
			       ? wc[v1].acode : wc[v2].acode);
		if (val <= wc[v1].acode)
			return(wc[v1].acode % 1000);
		else if (val <= wc[v2].acode)
			return(wc[v2].acode % 1000);
		else
			return(-1);
	}
	else
		return(-1);
}


/*
	Utility Routines
*/

/*
	Routine to tell if an item is being carried.
*/
int toting(int item)
{
	return(place[item] == -1);
}

/*
	Routine to tell if an item is present.
*/
int here (int item)
{
	return(place[item] == loc || toting(item));
}

/*
	Routine to test for darkness
*/
int dark()
{
	return(!(cond[loc] & LIGHT) &&
		(!prop[LAMP] ||
		!here(LAMP)));
}

/*
	Routine to tell if a location causes
	a forced move.
*/
int forced(int atloc)
{
	return(cond[atloc] == 2);
}

/*
	Routine true x% of the time.
*/
int pct(int x)
{
	return(random() % 100 < x);
}

/*
	Routine to tell if player is on
	either side of a two sided object.
*/
int at(int item)
{
	return(place[item] == loc || fixed[item] == loc);
}

/*
	Routine to carry an object
*/
void carry(int obj, int where)
{
	if (obj<MAXOBJ){
		if (place[obj] == -1)
			return;
		place[obj]=-1;
		++holding;
	}
}

/*
	Routine to drop an object
*/
void drop(int obj, int where)
{
	if (obj<MAXOBJ) {
		if (place[obj] == -1)
			--holding;
		place[obj]=where;
	}
	else
		fixed[obj-MAXOBJ]=where;
}

/*
	Routine to move an object
*/
void move(int obj, int where)
{
	int	from;

	from = (obj<MAXOBJ) ? place[obj] : fixed[obj];
	if (from>0 && from<=300)
		carry(obj, from);
	drop(obj, where);
}

/*
	Routine to destroy an object
*/
void dstroy(int obj)
{
	move(obj, 0);
}

/*
	Juggle an object
	currently a no-op
*/
void juggle(int loc)
{
}

/*
	routine to move an object and return a
	value used to set the negated prop values
	for the repository.
*/
int put(int obj, int where, int pval)
{
	move(obj, where);
	return((-1)-pval);
}
/*
	Routine to check for presence
	of dwarves..
*/
int dcheck()
{
	int	i;

	for (i =1; i < (DWARFMAX-1); ++i)
		if (dloc[i] == loc)
			return(i);
	return(0);
}

/*
	Convert  0 to WATER
		 1 to nothing
		 2 to OIL
*/
static int liq2 (int pbottle)
{
	return((1-pbottle)*WATER+(pbottle>>1)*(WATER+OIL));
}

/*
	Determine liquid in the bottle
*/
int liq()
{
	int	i, j;
	i=prop[BOTTLE];
	j=-1-i;
	return(liq2(i>j ? i : j));
}

/*
	Determine liquid at a location
*/
int liqloc(int loc)
{
	if (cond[loc]&LIQUID)
		return(liq2(cond[loc]&WATOIL));
	else
		return(liq2(1));
}

/*
	Fatal error routine
*/
void bug (int n)
{
	printf("Fatal error number %d\n", n);
	exit(1);
}


